/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "detail/wren_types.hpp"
#include "detail/strings_in_vector.hpp"
#include "detail/construct_foreign_class.hpp"
#include "detail/wren_class_name_from_type.hpp"
#include "detail/type_id.hpp"
#include <unordered_map>
#include <string>
#include <type_traits>
#include <functional>
#include <utility>

namespace wren {
	namespace detail {
		class ClassNameBase {
		public:
			ClassNameBase (const std::string_view& module_name, const std::string_view& class_name);
			ClassNameBase (const ClassNameBase& other) = default;

			std::size_t module_hash() const noexcept { return m_module_hash; }
			std::size_t class_hash() const noexcept { return m_class_name_hash; }

			bool operator== (const ClassNameBase& other) const;

		private:
			std::size_t m_module_hash;
			std::size_t m_class_name_hash;
		};

		class ClassNameOwning : StringsInVector<2>, public ClassNameBase {
		public:
			ClassNameOwning (const std::string_view& module_name, const std::string_view& class_name);

			std::string_view module_name() const { return this->as_string_view<0>(); }
			std::string_view class_name() const { return this->as_string_view<1>(); }
			operator ModuleAndName() const;

			using StringsInVector<2>::operator<;
			bool operator== (const ClassNameOwning& other) const;
		};

		class TempClassName : public ClassNameBase {
		public:
			TempClassName (std::string_view module_name, std::string_view class_name);
			TempClassName (const ClassNameOwning& model);
			bool operator== (const TempClassName& other) const;

			std::string_view module_name() const { return m_module_name; }
			std::string_view class_name() const { return m_class_name; }

		private:
			std::string_view m_module_name, m_class_name;
		};

		struct ClassNameEqual {
			using is_transparent = std::true_type;
			bool operator()(TempClassName left, TempClassName right) const;
		};

		struct ClassNameHash {
			using is_transparent = std::true_type;
			std::size_t operator()(TempClassName value) const;
		};

		class TypeIDHash {
		public:
			std::size_t operator()(TypeID tid) const;
		};

		define_has_typedef(class_type, ClassType);
	} //namespace detail

	//Customisation point - specialise for your own types if needed
	template <typename T>
	struct ClassMakerTraits {
	};

	template <typename T, typename... Args>
	struct ClassMakerTraits<detail::MakeForeignClass<T, Args...>> {
		typedef T class_type;
	};

	class ClassManager {
		typedef std::function<foreign_class_t()> make_foreign_class_t;
		typedef std::unordered_map<detail::ClassNameOwning, make_foreign_class_t, detail::ClassNameHash, detail::ClassNameEqual> ClassMap;
		typedef std::unordered_map<TypeID, const detail::ClassNameOwning*, detail::TypeIDHash> TypeMap;
	public:
		ClassManager();
		~ClassManager() noexcept;

		template <typename F>
		ClassManager& add_class_maker (std::string module_name, std::string_view class_name, F&& maker);
		foreign_class_t make_class(std::string_view module_name, std::string_view class_name);
		ModuleAndName wren_class_name_from_hash(TypeID hash) const;
		template <typename T>
		ModuleAndName wren_class_name_from_type() const;

	private:
		ClassMap::const_iterator add_class_maker_implem (
			std::string_view module_name,
			std::string_view class_name,
			make_foreign_class_t
		);

		void add_type (TypeID hash, const detail::ClassNameOwning* name);

		ClassMap m_classes;
		TypeMap m_types; //refers to memory in m_classes
	};

	template <typename F>
	inline ClassManager& ClassManager::add_class_maker (std::string module_name, std::string_view class_name, F&& maker) {
		typedef ClassMakerTraits<std::decay_t<F>> MakerTraits;

		auto it_new = this->add_class_maker_implem(module_name, class_name, std::forward<F>(maker));

		//See if we can deduce the class type that maker is supposed to work
		//with. This is not guaranteed to be an existing piece of info because
		//user code can provide their own function that doesn't use any actual
		//c++ class, just normal functions.
		if constexpr (detail::HasClassTypeTypedef<MakerTraits>::value) {
			const auto& owning_name = it_new->first;
			this->add_type(type_id<typename MakerTraits::class_type>(), &owning_name);
		}

		return *this;
	}

	template <typename T>
	inline ModuleAndName ClassManager::wren_class_name_from_type() const {
		return detail::wren_class_name_from_type<T>(*this);
	}
} //namespace wren
