/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <cstdint>
#include <cstddef>
#include <type_traits>
#include <utility>

namespace wren {
	namespace detail {
		[[gnu::const]]
		std::uint32_t runtime_crc32c (const char* data, std::size_t size, std::uint32_t crc);

		constexpr std::uint32_t g_castagnoli_polynomial = 0x1EDC6F41;

		[[gnu::const]]
		constexpr std::uint8_t reverse (std::uint8_t b) {
			//see https://graphics.stanford.edu/~seander/bithacks.html#ReverseByteWith64Bits
			return ((b * 0x80200802ULL) & 0x0884422110ULL) * 0x0101010101ULL >> 32;
		}

		[[gnu::const]]
		constexpr std::uint32_t reverse (std::uint32_t val) {
			return (reverse(static_cast<std::uint8_t>(val & 0xff)) << 24) |
				(reverse(static_cast<std::uint8_t>(val >> 8 & 0xff)) << 16) |
				(reverse(static_cast<std::uint8_t>(val >> 16 & 0xff)) << 8) |
				reverse(static_cast<std::uint8_t>(val >> 24 & 0xff));
		}

		consteval std::uint32_t calc_table_entry (std::uint32_t index, std::uint32_t polynomial) {
			polynomial = reverse(polynomial);
			auto c = static_cast<std::uint_fast32_t>(index);
			for (std::uint_fast32_t k = 0; k < 8; ++k) {
				c = (c >> 1) ^ ((c & 0x1u) ? polynomial : 0);
			}
			return static_cast<std::uint32_t>(c);
		}

		template <std::uint32_t Polynomial, typename T>
		struct PolynomialTableBase;

		template <std::uint32_t Polynomial, std::uint32_t... Indices>
		struct PolynomialTableBase<Polynomial, std::integer_sequence<std::uint32_t, Indices...>> {
			static constexpr std::size_t table_size = sizeof...(Indices);
			static constexpr std::uint32_t table[table_size] = {calc_table_entry(Indices, Polynomial)...};
		};

		template <std::uint32_t Polynomial>
		struct PolynomialTable : public PolynomialTableBase<Polynomial, decltype(std::make_integer_sequence<std::uint32_t, 256>{})> {
		};

		template <std::uint32_t Polynomial>
		constexpr inline auto g_polynomial_table = PolynomialTable<Polynomial>::table;

		template <std::uint32_t Polynomial, std::uint32_t XorIn=0xffffffff, std::uint32_t XorOut=XorIn>
		constexpr std::uint32_t crc32 (const char* data, std::size_t len, std::uint32_t crc) noexcept {
			//static_assert(g_polynomial_table<Polynomial>[0b10000000] == Polynomial);
			crc ^= XorIn;
			for (std::size_t z = 0; z < len; ++z) {
				crc = g_polynomial_table<Polynomial>[(crc ^ static_cast<unsigned char>(data[z])) & 0xff] ^ (crc >> 8);
			}
			return crc ^ XorOut;
		}
	} //namespace detail

	[[gnu::const]]
	constexpr std::uint32_t crc32c (const char* data, std::size_t size) {
		if (not std::is_constant_evaluated()) {
			return detail::runtime_crc32c(data, size, 0);
		}
		else {
			return detail::crc32<detail::g_castagnoli_polynomial>(data, size, 0);
		}
	}

	template <typename T, std::size_t N>
	[[gnu::const]]
	constexpr std::uint32_t crc32c (const T (&data)[N]) {
		return crc32c(static_cast<const char*>(data), N * sizeof(T));
	}
} //namespace wren
