/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "string_bt.hpp"
#include "crc32.hpp"
#include <string_view>
#include <cstdint>
#include <cassert>
#include <type_traits>

namespace wren {
	typedef decltype(crc32c(nullptr, 0)) ModuleAndNameHashType;
	class ModuleAndName;

	namespace detail {
		template <dhandy::bt::string Str>
		struct ModuleAndNameStaticStorage {
			static constexpr dhandy::bt::string value = Str;
		};

		[[gnu::const]]
		bool deep_equal (const ModuleAndName& a, const ModuleAndName& b) noexcept;
	} //unnamed namespace

	//Non-owning, lightweight class for storing a module name + generic name
	//pair. It holds a hash value too, which is currently only meaningful for
	//module name/class name pairs that are mapped to a c++ class.
	//Names are stored as consecutive strings within a single buffer with a
	//null-char in-between and at the end so that each name is null-terminated
	//and can be used with C APIs.
	//Keep the size small, ideally within 2 registers size. Altering this might
	//require changes to the assembly source files for DynafuncMaker.
	//When passed as a parameter by value, objects of ModuleAndName fit into
	//2 registers on 64 bit architectures. This is important because objects
	//of this class get hardcoded in the dynamic functions that DynafuncMaker
	//spits out.
	class ModuleAndName {
		friend bool detail::deep_equal(const ModuleAndName&, const ModuleAndName&) noexcept;
		template <dhandy::bt::string S1, dhandy::bt::string S2>
		friend consteval ModuleAndName make_module_and_name() noexcept;
	public:
		constexpr ModuleAndName();
		constexpr ModuleAndName (const char* base, std::size_t mod_name_len, std::size_t class_name_len);

		constexpr bool operator==(const ModuleAndName& other) const noexcept;
		constexpr bool operator!=(const ModuleAndName& other) const noexcept;
		constexpr bool deep_equal(const ModuleAndName& other) const noexcept;

		constexpr const char* module_name_char() const noexcept { return m_base; }
		constexpr const char* class_name_char() const noexcept { return m_base + 1 + m_mod_name_len; }
		constexpr std::size_t module_name_len() const noexcept { return m_mod_name_len; }
		constexpr std::size_t class_name_len() const noexcept { return m_class_name_len; }
		constexpr std::string_view module_name() const noexcept { return {module_name_char(), m_mod_name_len}; }
		constexpr std::string_view class_name() const noexcept { return {class_name_char(), m_class_name_len}; }
		constexpr const char* buffer_address() const noexcept { return m_base; }

	private:
		constexpr ModuleAndName (const char* base, ModuleAndNameHashType hash, std::size_t mod_name_len, std::size_t class_name_len);

		constexpr std::size_t raw_buffer_size() const noexcept;

		const char* m_base;
		ModuleAndNameHashType m_hash;
		std::uint16_t m_mod_name_len;
		std::uint16_t m_class_name_len;

		static_assert(
			sizeof(m_hash) <= sizeof(std::uint32_t),
			"Hash type is too large, review this struct or change it back to uint32_t"
		);
	};

	union RawAccessModuleAndName {
		constexpr RawAccessModuleAndName (ModuleAndName mn) :
			module_and_name(mn)
		{ }

		ModuleAndName module_and_name;

		struct {
			void* string;
			void* data;
		} raw;

		static_assert(sizeof(raw) == sizeof(module_and_name));
	};

	constexpr ModuleAndName::ModuleAndName() :
		ModuleAndName(nullptr, 0, 0, 0)
	{}

	constexpr ModuleAndName::ModuleAndName (const char* base, std::size_t mod_name_len, std::size_t class_name_len) :
		ModuleAndName(base, crc32c(base, mod_name_len + class_name_len + 2), mod_name_len, class_name_len)
	{}

	constexpr ModuleAndName::ModuleAndName (const char* base, ModuleAndNameHashType hash, std::size_t mod_name_len, std::size_t class_name_len) :
		m_base(base),
		m_hash(hash),
		m_mod_name_len(mod_name_len),
		m_class_name_len(class_name_len)
	{}

	constexpr bool ModuleAndName::operator==(const ModuleAndName& other) const noexcept {
		const bool retval = (this->m_hash == other.m_hash and deep_equal(other));
		return retval;
	}

	constexpr bool ModuleAndName::operator!=(const ModuleAndName& other) const noexcept {
		return not this->operator==(other);
	}

	constexpr bool ModuleAndName::deep_equal(const ModuleAndName& other) const noexcept {
		if (not std::is_constant_evaluated()) {
			return detail::deep_equal(*this, other);
		}
		else {
			if (this->raw_buffer_size() != other.raw_buffer_size())
				return false;
			for (std::size_t z = 0; z < raw_buffer_size(); ++z) {
				if (this->m_base[z] != other.m_base[z])
					return false;
			}
			return true;
		}
	}

	constexpr std::size_t ModuleAndName::raw_buffer_size() const noexcept {
		return m_mod_name_len + m_class_name_len + 2;
	}

	template <dhandy::bt::string S1, dhandy::bt::string S2>
	consteval ModuleAndName make_module_and_name() noexcept {
		using dhandy::bt::string;
		using detail::ModuleAndNameStaticStorage;
		constexpr string null_char{"\0"};
		using StaticStorage = ModuleAndNameStaticStorage<S1 + null_char + S2>;

		constexpr const char* data = StaticStorage::value.data();
		constexpr std::uint16_t s1_len = static_cast<std::uint16_t>(S1.size());
		constexpr std::uint16_t s2_len = static_cast<std::uint16_t>(S2.size());
		constexpr ModuleAndNameHashType hash = crc32c(
			StaticStorage::value.data(),
			StaticStorage::value.size()
		);

		return ModuleAndName{data, hash, s1_len, s2_len};
	}

	//short-hand alias for make_module_and_name
	template <dhandy::bt::string S1, dhandy::bt::string S2>
	constexpr ModuleAndName MN = make_module_and_name<S1, S2>();
} //namespace wren
