/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "wrenpp/detail/crc32.hpp"
#include "crc32_config.h"

namespace wren::detail {
#if HAVE_NEON
[[gnu::const]] bool has_crc32_neon() noexcept;
[[gnu::pure]] std::uint32_t crc32c_neon(const char*, std::size_t, std::uint32_t) noexcept;
#endif
#if HAVE_SSE42
[[gnu::const]] bool has_crc32_sse42() noexcept;
[[gnu::pure]] std::uint32_t crc32c_sse42(const char*, std::size_t, std::uint32_t) noexcept;
#endif
} //namespace wren::detail

namespace wren::detail {
	//Some useful polynomials:
	// zlib: 0x04C11DB7
	// castagnoli (intel):  0x1EDC6F41

	namespace {
		[[gnu::const]]
		auto best_crc32_function() {
#if HAVE_NEON
			if (has_crc32_neon())
				return &crc32c_neon;
#endif

#if HAVE_SSE42
			if (has_crc32_sse42())
				return &crc32c_sse42;
#endif

			return &crc32<detail::g_castagnoli_polynomial>;
		}
	} //unnamed namespace

	[[gnu::const]]
	std::uint32_t runtime_crc32c (const char* data, std::size_t size, std::uint32_t crc) {
		static const auto crc32c_implem = best_crc32_function();
		return (*crc32c_implem)(data, size, crc);
	}
} //namespace wren::detail
