/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <smmintrin.h>
#include <immintrin.h>
#include <cstdint>

// Byte-boundary alignment issues
#define CALC_CRC(op, crc, type, buf, len)							   \
  do {																  \
	for (; (len) >= sizeof (type); (len) -= sizeof(type), buf += sizeof (type)) { \
	  (crc) = op((crc), *(type *) (buf));							   \
	}																   \
  } while(0)

namespace wren::detail {
	[[gnu::const]]
	bool has_crc32_sse42() noexcept {
		const bool has_hw_crc32 = (__builtin_cpu_supports("sse4.2") ? true : false);
		return has_hw_crc32;
	}

	/* Compute CRC-32C using the Intel hardware instruction. */
	/* for better parallelization with bigger buffers see
	   http://www.drdobbs.com/parallel/fast-parallelized-crc-computation-using/229401411 */
	[[gnu::pure]]
	std::uint32_t crc32c_sse42(const char* input, std::size_t len, std::uint32_t crc) noexcept {
		//see https://github.com/rurban/smhasher/blob/master/crc32_hw.c
		constexpr std::size_t align_size = alignof(std::uint64_t);
		constexpr std::size_t align_mask = align_size - 1;

		// XOR the initial CRC with INT_MAX
		//crc ^= 0xFFFFFFFF;
		crc = ~crc;

		// Align the input to the word boundary
		for (; (len > 0) && (reinterpret_cast<std::uintptr_t>(input) & align_mask); len--, input++) {
			crc = _mm_crc32_u8(crc, *input);
		}

		// Blast off the CRC32 calculation
#if defined(__x86_64__) || defined(__aarch64__)
		CALC_CRC(_mm_crc32_u64, crc, std::uint64_t, input, len);
#endif
		CALC_CRC(_mm_crc32_u32, crc, std::uint32_t, input, len);
		CALC_CRC(_mm_crc32_u16, crc, std::uint16_t, input, len);
		CALC_CRC(_mm_crc32_u8, crc, std::uint8_t, input, len);

		// Post-process the crc
		return ~crc;
	}
} //namespace wren::detail
